import sys, threading
import ConfigParser
import paramiko

####################################################################
## Name: mongo_db_cluster_locker 
## - A python module meant for consistent backup of MongoDB Shard Clusters
## It is meant to be run from Cloud Protection Manager (CPM) Server, and 
## to connect remotely to the MongoDB cluster members using SSH
## It des so with the following steps:
## 1) stop the load balancer and wait out any ongoing migrations process
## 2) Flush and lock all MongoDB shard servers in parallel
## 3) Stop the MongoDB service in one of the clusters config servers
## 
## Interface: class MongoCluster:
## declaration cluster = MongoCluster (<path to configuration file>)
## Then cluster.lock () and cluster.unlock ()
##
## Written by Uri Wolloch - info@n2ws.com
## September 2013
## License: Anyone may use. moify & distribute freely
##          Please leave the in-code documentation
##
## This code is provided as-is without any guarantee or warranty. 
## Please test carefully before using in production environment.
##
## N2W Software - www.n2ws.com
####################################################################

class RemoteServer (object):
    """
    Base class containing connection details for a remote
    MongoDB server, connected to usin SSH
    """
    address = None
    ssh_port = 22
    ssh_user = None
    ssh_password = None
    ssh_key_file = None

    mongo_port = 27027
    mongo_auth = False
    mongo_user = None
    mongo_password = None
    mongo_db = 'admin'

    timeout_secs = 30
    
    def read_config (self, parser, section):
    
        # ssh details
        self.address = parser.get (section, 'address')
        if parser.has_option(section, 'ssh_port'):
            self.ssh_port = parser.getint (section, 'ssh_port')
        self.ssh_user = parser.get (section, 'ssh_user')
        if parser.has_option(section, 'ssh_password'):
            self.ssh_password = parser.get (section, 'ssh_password')
        if parser.has_option(section, 'ssh_key_file'):
            self.ssh_key_file = parser.get (section, 'ssh_key_file')
        # mongo details, not all types need it
        if parser.has_option (section, 'mongo_port'):
            self.mongo_port = parser.getint (section, 'mongo_port')
        if parser.has_option(section, 'mongo_user'):
            self.mongo_auth = True
            self.mongo_user = parser.get (section, 'mongo_user')
            self.mongo_password = parser.get (section, 'mongo_password')
        if parser.has_option (section, 'timeout_secs'):
            self.timeout_secs = parser.getint (section, 'timeout_secs')
            
    def connect_and_exec (self, command):
        """
        Connects to a remote maching using SSH, and executing the "command"
        """
        if self.ssh_key_file and len (self.ssh_key_file) > 0:
            try:
                key = paramiko.RSAKey.from_private_key_file(self.ssh_key_file)
            except:
                sys.stderr.write ('SSH Key file %s non-existent or bad\n' % self.ssh_key_file)
                return False
        else:
            key = None
            
        try:
            transport = paramiko.Transport((self.address,self.ssh_port))
            transport.connect(username=self.ssh_user, pkey=key)
            channel = transport.open_session()
            channel.settimeout(self.timeout_secs)
            channel.exec_command (command)
        except paramiko.SSHException, e:
            sys.stderr.write (e.message+'\n')
            return False
        except Exception, e:
            sys.stderr.write ('SSH Connectivity Issue: %s\n', e)
            return False
 
        # capture stderr output from the remote host and output in the local
        stderr_lines = channel.makefile_stderr('rb', -1).readlines()
        for line in stderr_lines:
            sys.stderr.write (line)
           
        retval = channel.recv_exit_status()
       
        return retval==0
    
    def prepare_mongo_command_line (self, command):
    
        # construct remote command line to be executed on the remote mongodb server
        if self.mongo_user and len(self.mongo_user):
            remote_cmd = 'mongo localhost:%d/%s -u %s -p %s -eval \'%s\'' % \
                (self.mongo_port, self.mongo_db, self.mongo_user, self.mongo_password, command)
        else: # no auth
            remote_cmd = 'mongo localhost:%d/%s -eval \'%s\'' % \
                (self.mongo_port, self.mongo_db, command)
        return remote_cmd
    
class ShardServer (RemoteServer):
    """
    Class representing a MongoDB shard server 
    (or secondary member of shard replication set)
    There can be any number of shard servers in sections SHARDSERVER_1....x
    """

    operation_succeded = False
    def lock (self):
        self.operation_succeded = False
        retval = self.connect_and_exec (self.prepare_mongo_command_line ('db.fsyncLock();'))
        self.operation_succeded = retval
    def unlock (self):
        self.operation_succeded = False
        retval = self.connect_and_exec (self.prepare_mongo_command_line ('db.fsyncUnlock();'))
        self.operation_succeded = retval
    
class ConfigServer (RemoteServer):
    """
    Class representing a MongoDB config server 
    Only 1 should be present in the script configuration file
    """
    mongodb_service_start_command = 'sudo service mongodb start'
    mongodb_service_stop_command = 'sudo service mongodb stop'
    def start_mongod_service (self):
        return self.connect_and_exec (self.mongodb_service_start_command)
    def stop_mongod_service (self):
        return self.connect_and_exec (self.mongodb_service_stop_command)
        
    def read_config (self, parser, section):
        super(ConfigServer, self).read_config (parser, section)
        
        if parser.has_option (section, 'mongo_db_service_start'):
            self.mongodb_service_start_command = parser.get (section, 'mongo_db_service_start')
        if parser.has_option (section, 'mongo_db_service_stop'):
            self.mongodb_service_stop_command = parser.get (section, 'mongo_db_service_stop')
        
        
class MongoSServer (RemoteServer):
    """
    Class representing a MongoS server 
    Only 1 should be present in the script configuration file
    Used to stop and start the load balancer
    """
    mongo_db = 'config'
    
    def stop_balancer (self):
        return self.connect_and_exec (self.prepare_mongo_command_line ('sh.setBalancerState(false); '
                'while( sh.isBalancerRunning() ) { '
                'sleep(1000); '
                '}'))
    def start_balancer (self):
        return self.connect_and_exec (self.prepare_mongo_command_line ('sh.setBalancerState(true);'))

class MongoCluster (object):
    """
    This is the main class: represents a whole MongoDB cluster
    Contains a MongoS server, a config server and multiple shard servers
    It creates all contained objects
    """

    mongos = MongoSServer ()
    config_server = ConfigServer ()
    shard_servers = []
    
    def __init__ (self, config_file_path='mongo_db_cluster_locker.cfg'):
    
        parser = ConfigParser.RawConfigParser()
        parser.read (config_file_path)
        
        self.mongos.read_config (parser, 'MONGOS')
        self.config_server.read_config (parser, 'CONFIGSERVER')
        
        # create shard server objects and read their configuration
        for section in parser.sections ():
            if section.startswith ('SHARDSERVER'):
                shard_server = ShardServer ()
                shard_server.read_config (parser, section)
                self.shard_servers.append (shard_server)
                
                
    def lock (self):
        """
        Locks the mongo cluster before snapshots are taken
        Makes sure everything works and returns True
        If some of the locking fails, this function will
        rollback all other changes and return False
        All output is to stderr
        """
        # According to MongoDB documentation, the first thing we need to do
        # is stop the balancer and make sure no migrtions are happening
        retval = self.mongos.stop_balancer ()
        if retval:
           sys.stderr.write ('Load balancer stopped successfully\n')
        else:
           sys.stderr.write ('Load balancer failed to stop, aborting\n')
           return False
          
        # lock all shard servers in parallel - using separate threads  
        shard_lock_threads = []
        for shard_server in self.shard_servers:
            thread = threading.Thread(target=shard_server.lock ())
            shard_lock_threads.append (thread)
            thread.start ()
            
        # Waiting for threading completion
        for thread in shard_lock_threads:
            thread.join ()
           
        # now make sure they all succeeded
        all_succeeded = True
        for shard_server in self.shard_servers:
            if shard_server.operation_succeded:
                sys.stderr.write ('Lock of shard server %s succeeded\n' % shard_server.address)
            else:
                all_succeeded = False
                sys.stderr.write ('Lock of shard server %s failed\n' % shard_server.address)
                
        if all_succeeded:
            sys.stderr.write ('All shard servers locked successfully\n')
        else:
            sys.stderr.write ('Some shard servers failed to lock, undoing and aborting\n')
            shard_lock_threads = []
            shard_lock_threads = []
            for shard_server in self.shard_servers:
                if shard_server.operation_succeded:
                    thread = threading.Thread(target=shard_server.unlock ())
                    shard_lock_threads.append (thread)
                    thread.start ()
                
            # Waiting for threading completion
            for thread in shard_lock_threads:
                thread.join ()
            
            self.mongos.start_balancer ()
            return False
        
        # Last stage, stop the mongod service  in the config server,
        # will stop configuration changes from happening
        retval = self.config_server.stop_mongod_service ()
        if retval:
            sys.stderr.write ('Config server stopped successfully\n')
        else:
            sys.stderr.write ('Failed to stop config server, undoing and aborting\n')
            shard_lock_threads = []
            for shard_server in self.shard_servers:
                thread = threading.Thread(target=shard_server.unlock ())
                shard_lock_threads.append (thread)
                thread.start ()
                
            # Waiting for threading completion
            for thread in shard_lock_threads:
                thread.join ()
            
            self.mongos.start_balancer ()
            return False
            
        sys.stderr.write ('MongoDB cluster lock completed successfully!\n')
        return True
        
        
    def unlock (self):
        """
        unlocks the mongo cluster after snapshots were taken
        Makes sure everything works and returns True
        If some of the locking fails, this function will produce error msg
        All output is to stderr
        """
        all_successful = True
        
        # we'll perform the operation in reverse order than the locking function
        retval = self.config_server.start_mongod_service ()
        if retval:
            sys.stderr.write ('Config server started successfully\n')
        else:
            all_successful = False
            sys.stderr.write ('Failed to start config server\n')
        
        # Now unlock shard servers
        shard_unlock_threads = []
        for shard_server in self.shard_servers:
            thread = threading.Thread(target=shard_server.unlock ())
            shard_unlock_threads.append (thread)
            thread.start ()
            
        # Waiting for threading completion
        for thread in shard_unlock_threads:
            thread.join ()
            
        # now check success
        for shard_server in self.shard_servers:
            if shard_server.operation_succeded:
                sys.stderr.write ('Shard server %s unlocked successfully\n' % shard_server.address)
            else:
                all_successful = False
                sys.stderr.write ('Unlock of shard server %s failed\n' % shard_server.address)
        
        # now, last stage, start load balancer again
        retval = self.mongos.start_balancer ()
        if retval:
            sys.stderr.write ('Load balancer started successfully\n')
        else:
            all_successful = False
            sys.stderr.write ('Couldn\'t start load balancer\n')
            
        return all_successful
        
                    
        
        
        
        

        

